# Check the following urls for more info about Pan-STARRS:
#
#     https://outerspace.stsci.edu/display/PANSTARRS/PS1+Image+Cutout+Service#PS1ImageCutoutService-ImportantFITSimageformat,WCS,andflux-scalingnotes
#     https://outerspace.stsci.edu/display/PANSTARRS/PS1+Stack+images#PS1Stackimages-Photometriccalibration
#
# For DES:
#
#     https://des.ncsa.illinois.edu/releases/dr1/dr1-docs/processing
#
# For SDSS:
#
#     https://www.sdss.org/dr12/algorithms/fluxcal/#SDSStoAB
#     https://data.sdss.org/datamodel/files/BOSS_PHOTOOBJ/frames/RERUN/RUN/CAMCOL/frame.html
#
# Some parts of this notebook are based on https://github.com/djones1040/PS1_surface_brightness/blob/master/Surface%20Brightness%20Tutorial.ipynb and codes from Lluís Galbany

import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import aplpy

plt.rcParams['mathtext.fontset'] = "cm"

import sep
from photutils.aperture import aperture_photometry, CircularAperture

from astropy.io import fits
from astropy import units as u, wcs
from astropy.cosmology import FlatLambdaCDM

from hostphot._constants import workdir, font_family
from hostphot.utils import (
    get_survey_filters,
    check_survey_validity,
    check_filters_validity,
    calc_sky_unc,
    get_image_exptime,
    survey_pixel_scale,
    check_work_dir,
    magnitude_calculation,
    suppress_stdout,
    bkg_surveys,
)
from hostphot.image_cleaning import remove_nan
from hostphot.dust import calc_extinction

import warnings
from astropy.utils.exceptions import AstropyWarning

H0 = 70
Om0 = 0.3
__cosmo__ = FlatLambdaCDM(H0, Om0)
sep.set_sub_object_limit(1e4)

# ----------------------------------------
def choose_cosmology(cosmo):
    """Updates the cosmology used to calculate the aperture size.

    Parameters
    ----------
    cosmo: `astropy.cosmology` object
        Cosmological model. E.g. :func:`FlatLambdaCDM(70, 0.3)`.
    """
    global __cosmo__
    __cosmo__ = cosmo


# -------------------------------
def calc_aperture_size(z, ap_radius):
    """Calculates the size of the aperture in arsec,
    for aperture photometry, given a physical size.

    Parameters
    ----------
    z: float
        Redshift.
    ap_radius: float
        Physical aperture size in kpc.

    Returns
    -------
    radius_arcsec: float
        Aperture size in arcsec.
    """
    ap_radius = ap_radius * u.kpc

    # transverse separations
    transv_sep_per_arcmin = __cosmo__.kpc_proper_per_arcmin(z)
    transv_sep_per_arcsec = transv_sep_per_arcmin.to(u.kpc / u.arcsec)

    radius_arcsec = ap_radius / transv_sep_per_arcsec

    return radius_arcsec.value


def extract_aperture_flux(data, error, px, py, radius):
    """Extracts aperture photometry of a single image.

    Parameters
    ----------
    data: array
        Image data in a 2D numpy array.
    error: array
        Errors of ``data``.
    px: float
        x-axis pixel coordinate of the aperture center.
    py: float
        y-axis pixel coordinate of the aperture center.
    radius: float
        Aperture radius in pixels.

    Returns
    -------
    raw_flux: float
        Aperture photometry ("raw" flux).
    raw_flux_err: float
        Uncertainty on the aperture photometry.
    """
    aperture = CircularAperture((px, py), r=radius)
    ap_results = aperture_photometry(data, aperture, error=error)
    raw_flux = ap_results["aperture_sum"][0]
    raw_flux_err = ap_results["aperture_sum_err"][0]

    return raw_flux, raw_flux_err


def plot_aperture(hdu, px, py, radius_pix, title, outfile=None):
    """Plots the aperture for the given parameters.

    Parameters
    ----------
    data: ndarray
        Data of an image.
    px: float
        X-axis center of the aperture in pixels.
    py: float
        Y-axis center of the aperture in pixels.
    radius_pix: float
        Aperture radius in pixels.
    img_wcs: WCS
        Image's WCS.
    title: str
        Title of the figure.
    outfile: str, default ``None``
        If given, path where to save the output figure.
    """

    figure = plt.figure(figsize=(10, 10))
    title, label = title.split('|')

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        fig = aplpy.FITSFigure(hdu, figure=figure)

    with suppress_stdout():
        fig.show_grayscale(stretch='arcsinh')

    # plot SN and aperture
    fig.show_markers(px, py, edgecolor='k', facecolor='aqua', 
                     marker='*', s=200, label='SN', coords_frame='pixel')
    fig.show_circles(px, py, radius_pix, coords_frame='pixel', linewidth=3, edgecolor='r')
    fig.show_lines([], [], color='r', lw=3, label=label)  # for the legend only

    #ticks
    fig.tick_labels.set_font(**{'family':font_family, 'size':18})
    fig.tick_labels.set_xformat('dd.dd')
    fig.tick_labels.set_yformat('dd.dd')
    fig.ticks.set_length(6)

    fig.axis_labels.set_font(**{'family':font_family, 'size':18})

    fig.set_title(title, **{'family':font_family, 'size':24})
    fig.set_theme('publication')
    fig.ax.legend(fancybox=True, framealpha=1, prop={'size':20, 'family':font_family})

    if outfile:
        plt.savefig(outfile, bbox_inches="tight")
        plt.close(figure)
    else:
        plt.show()

def photometry(
    name,
    ra,
    dec,
    z,
    filt,
    survey,
    ap_radii=1,
    bkg_sub=None,
    use_mask=True,
    correct_extinction=True,
    save_plots=True,
):
    """Calculates the local aperture photometry in a given radius.

    Parameters
    ----------
    name: str
        Name of the object to find the path of the fits file.
    ra: float
        Right Ascensions in degrees to center the aperture.
    dec: float
        Declinations in degrees to center the aperture.
    z: float
        Redshift of the object to estimate the physical calculate
        of the aperture.
    filt: str
        Filter to use to load the fits file.
    survey: str
        Survey to use for the zero-points and pixel scale.
    ap_radii: float or list-like, default ``1``
        Physical size of the aperture in kpc.
    bkg_sub: bool, default ``None``
        If ``True``, the image gets background subtracted. By default, only
        the images that need it get background subtracted (WISE, 2MASS and
        VISTA).
    use_mask: bool, default ``True``
        If ``True``, the masked fits files are used. These must have
        been created beforehand.
    correct_extinction: bool, default `True`
        If `True`, corrects for Milky-Way extinction using the recalibrated dust maps
        by Schlafly & Finkbeiner (2011) and the extinction law from Fitzpatrick (1999).
    save_plots: bool, default ``True``
        If ``True``, the figure with the aperture is saved.

    Returns
    -------
    mags: list
        Aperture magnitudes for the given aperture radii.
    mags_err: list
        Aperture magnitude errors for the given aperture radii.
    fluxes: list
        Aperture flux for the given aperture radii.
    fluxes_err: list
        Aperture flux errors for the given aperture radii.
    zp: float
        Zeropoint.
    """
    if survey == 'SkyMapper':
        warnings.warn(("SkyMapper photometry is not completely trustworthy due to imprecision in "
                       "the zeropoint for extended sources, which might be solved in a future data release."))
        
    check_survey_validity(survey)
    check_work_dir(workdir)
    obj_dir = os.path.join(workdir, name)
    if use_mask:
        suffix = "masked_"
    else:
        suffix = ""
    fits_file = os.path.join(obj_dir, f"{suffix}{survey}_{filt}.fits")

    hdu = fits.open(fits_file)
    hdu = remove_nan(hdu)

    header = hdu[0].header
    data = hdu[0].data
    exptime = get_image_exptime(header, survey)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        img_wcs = wcs.WCS(header, naxis=2)

    data = data.astype(np.float64)
    bkg = sep.Background(data)
    bkg_rms = bkg.globalrms
    if (bkg_sub is None and survey in bkg_surveys) or bkg_sub is True:
        data_sub = np.copy(data - bkg)
    else:
        data_sub = np.copy(data)

    # turn float into a list
    if isinstance(ap_radii, (float, int)):
        ap_radii = [ap_radii]

    pixel_scale = survey_pixel_scale(survey, filt)
    if survey == 'UKIDSS' and filt == 'J':
        if 'LAS' in header['PROJECT'].split('/')[-1]:
            # if the J comes from the LAS then the observations were micro-stepped
            # in the J-band so the pixel size is smaller.
            nustep = header['NUSTEP']
            pixel_scale /= nustep/2  # divided by 2 for the two dimensions (x and y)

    # background error
    if survey in ["LegacySurvey"]:
        invvar_map = hdu[1].data
        error = np.sqrt(1/invvar_map)
    else:
        error = calc_sky_unc(data_sub, exptime)

    mags, mags_err = [], []
    fluxes, fluxes_err = [], []
    px, py = img_wcs.wcs_world2pix(ra, dec, 1)

    for ap_radius in ap_radii:
        # aperture photometry
        radius_arcsec = calc_aperture_size(z, ap_radius)
        radius_pix = radius_arcsec / pixel_scale

        flux, flux_err = extract_aperture_flux(
            data_sub, error, px, py, radius_pix
        )

        ap_area = np.pi * (radius_pix**2)
        mag, mag_err, flux, flux_err, zp = magnitude_calculation(
            flux,
            flux_err,
            survey,
            filt,
            ap_area,
            header,
            bkg_rms,
        )
        """
        if survey in ["LegacySurvey"]:
            invvar_map = hdu[1].data
            var_map = 1/invvar_map
            sum_var, _ = extract_aperture_flux(
                        var_map, error, px, py, radius_pix
                    )
            extra_flux_err = np.sqrt(sum_var) 
            flux_err = np.sqrt(flux_err**2 + extra_flux_err**2)

            extra_err = np.abs(2.5 * flux_err / (flux * np.log(10)))
            mag_err = np.sqrt(mag_err**2 + extra_err**2)
        """
        # extinction correction is optional
        if correct_extinction:
            A_ext = calc_extinction(filt, survey, ra, dec)
            mag -= A_ext
            flux *= 10**(0.4*A_ext)

        mags.append(mag)
        mags_err.append(mag_err)
        fluxes.append(flux)
        fluxes_err.append(flux_err)

        if save_plots:
            outfile = os.path.join(
                obj_dir, f"local_{survey}_{filt}_{ap_radius}kpc.jpg"
            )
            title = f'{name}: {survey}-${filt}$|r$={ap_radius}$ kpc @ $z={z}$'
            plot_aperture(hdu, px, py, radius_pix, title, outfile)

    return mags, mags_err, fluxes, fluxes_err, zp


def multi_band_phot(
    name,
    ra,
    dec,
    z,
    filters=None,
    survey="PS1",
    ap_radii=1,
    bkg_sub=None,
    use_mask=True,
    correct_extinction=True,
    save_plots=True,
    save_results=True,
    raise_exception=True,
):
    """Calculates the local aperture photometry for multiple filters.

    Parameters
    ----------
    name: str
        Name of the object to find the path of the fits file.
    ra: float
        Right Ascensions in degrees to center the aperture.
    dec: float
        Declinations in degrees to center the aperture.
    z: float
        Redshift of the object to estimate the physical calculate
        of the aperture.
    filters: str, default, ``None``
        Filters to use to load the fits files. If ``None`` use all
        the filters of the given survey.
    survey: str, default ``PS1``
        Survey to use for the zero-points and pixel scale.
    ap_radii: float or list-like, default ``1``
        Physical size of the aperture in kpc.
    bkg_sub: bool, default ``None``
        If ``True``, the image gets background subtracted. By default, only
        the images that need it get background subtracted (WISE, 2MASS and
        VISTA).
    use_mask: bool, default ``True``
        If ``True``, the masked fits files are used. These must have
        been created beforehand.
    correct_extinction: bool, default `True`
        If `True`, corrects for Milky-Way extinction using the recalibrated dust maps
        by Schlafly & Finkbeiner (2011) and the extinction law from Fitzpatrick (1999).
    save_plots: bool, default ``True``
        If ``True``, the figure with the aperture is saved.
    save_results: bool, default ``True``
        If ``True``, the magnitudes are saved into a csv file.
    raise_exception: bool, default ``True``
        If ``True``, an exception is raised if the photometry fails for any filter.

    Returns
    -------
    results_dict: dict
        Dictionary with the object's photometry and other info.

    Examples
    --------
    >>> import hostphot.local_photometry as lp
    >>> ap_radii = [3, 4]  # aperture radii in units of kpc
    >>> ra, dec =  308.22579, 9.92853 # coords of SN2004eo
    >>> z = 0.0157  # redshift
    >>> survey = 'PS1'
    >>> results = lp.multi_band_phot(name, ra, dec, z,
                            survey=survey, ap_radii=ap_radii,
                            use_mask=True, save_plots=True)
    """
    check_survey_validity(survey)
    if filters is None:
        if survey=='HST':
            raise ValueError("For HST, the filter needs to be specified!")
        filters = get_survey_filters(survey)
    else:
        check_filters_validity(filters, survey)
    if survey=='HST':
        filters = [filters]

    # turn float into a list
    if isinstance(ap_radii, (float, int)):
        ap_radii = [ap_radii]

    results_dict = {
        "name": name,
        "ra": ra,
        "dec": dec,
        "redshift": z,
        "survey": survey,
    }
    
    for filt in filters:
        try:
            mags, mags_err, fluxes, fluxes_err, zp = photometry(
                name,
                ra,
                dec,
                z,
                filt,
                survey,
                ap_radii,
                bkg_sub,
                use_mask,
                correct_extinction,
                save_plots,
            )
            for radius, mag, mag_err, flux, flux_err in zip(
                ap_radii, mags, mags_err, fluxes, fluxes_err
            ):
                results_dict[f"{filt}_{radius}"] = mag
                results_dict[f"{filt}_{radius}_err"] = mag_err
                results_dict[f"{filt}_{radius}_flux"] = flux
                results_dict[f"{filt}_{radius}_flux_err"] = flux_err
            results_dict[f"{filt}_zeropoint"] = zp
        except Exception as exc:
            if raise_exception is True:
                raise Exception(exc)
            else:
                for radius in ap_radii:
                    results_dict[f"{filt}_{radius}"] = np.nan
                    results_dict[f"{filt}_{radius}_err"] = np.nan
                    results_dict[f"{filt}_{radius}_flux"] = np.nan
                    results_dict[f"{filt}_{radius}_flux_err"] = np.nan
                results_dict[f"{filt}_zeropoint"] = np.nan

    if save_results is True:
        outfile = os.path.join(workdir, name, f"{survey}_local.csv")
        phot_df = pd.DataFrame(
            {key: [val] for key, val in results_dict.items()}
        )
        phot_df.to_csv(outfile, index=False)

    return results_dict
